// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "TrueTypeAsset.hpp"
#include "asset/AssetManager.hpp"

using namespace Dawn;

TrueTypeAsset::TrueTypeAsset(AssetManager *assMan, std::string name) :
  Asset(assMan, name),
  loader(name + ".truetype")
{
  this->locks.onLockRemoved = [&](usagelockid_t id){
    auto texture = this->textureByLock[id];

    assertNotNull(texture, "TrueTypeAsset::TrueTypeAsset: Texture cannot be null");

    std::vector<usagelockid_t> &lbt = this->locksByTexture[texture];
    auto it0 = std::find(lbt.begin(), lbt.end(), id);
    assertTrue(it0 != lbt.end(), "TrueTypeAsset::TrueTypeAsset: Could not remove locksByTexture[texture]");
    lbt.erase(it0);

    auto it1 = this->textureByLock.find(id);
    assertTrue(it1 != this->textureByLock.end(), "TrueTypeAsset::TrueTypeAsset: Could not remove textureByLock");
    this->textureByLock.erase(it1);

    if(lbt.empty()) {
      auto it2 = locksByTexture.find(texture);
      assertTrue(it2 != locksByTexture.end(), "TrueTypeAsset::TrueTypeAsset: Could not remove locksByTexture");
      locksByTexture.erase(it2);
      
      auto it3 = textureByStyle.begin();
      while(it3 != textureByStyle.end()) {
        if(it3->second == texture) it3 = textureByStyle.erase(it3);
        ++it3;
      }

      auto it4 = std::find(textures.begin(), textures.end(), texture);
      assertTrue(it4 != textures.end(), "TrueTypeAsset::TrueTypeAsset: Could not remove textureByStyle");
      textures.erase(it4);

      delete texture;
    }
  };
}

void TrueTypeAsset::updateSync() {

}

void TrueTypeAsset::updateAsync() {
  if(this->state != TRUE_TYPE_ASSET_STATE_INITIAL) return;

  this->state = TRUE_TYPE_ASSET_STATE_OPEN;
  this->loader.open();

  this->state = TRUE_TYPE_ASSET_STATE_VALIDATE_HEADER;

  uint8_t buffer[64];
  this->loader.rewind();
  size_t read = this->loader.read(buffer, sizeof(char) * 6);
  assertTrue(read == (6 * sizeof(char)), "TrueTypeAsset::updateAsync: Could not read header");
  buffer[6] = '\0';

  // Confirm "DE_TTF"
  assertTrue(std::string((char *)buffer) == "DE_TTF", "TrueTypeAsset::updateAsync: Header is invalid (Missing DE_TTF)");

  // Vertical bar
  this->loader.read(buffer, 1);
  assertTrue(buffer[0] == '|', "TrueTypeAsset::updateAsync: Header is invalid (Missing first vertical bar)");

  // Read version
  this->state = TRUE_TYPE_ASSET_STATE_VALIDATE_VERSION;
  read = this->loader.read(buffer, sizeof(char) * 5);
  assertTrue(buffer[0] == '3', "TrueTypeAsset::updateAsync: Version is invalid 3");
  assertTrue(buffer[1] == '.', "TrueTypeAsset::updateAsync: Version is invalid .");
  assertTrue(buffer[2] == '0', "TrueTypeAsset::updateAsync: Version is invalid 0(1)");
  assertTrue(buffer[3] == '0', "TrueTypeAsset::updateAsync: Version is invalid 0(2)");
  assertTrue(buffer[4] == '|', "TrueTypeAsset::updateAsync: Version is invalid (Missing second vertical bar)");

  // Read the count of font styles / variants.
  size_t styleListBegin = this->loader.getPosition();
  this->state = TRUE_TYPE_ASSET_STATE_READ_VARIANT_COUNT;
  read = this->loader.read(buffer, 64);
  assertTrue(read > 0, "TrueTypeAsset::updateAsync: Could not read variant count");

  // Get position of vertical bar.
  size_t i = 0;
  while(buffer[i] != '|' && i < 64) i++;
  assertTrue(buffer[i] == '|', "TrueTypeAsset::updateAsync: Could not find vertical bar");
  styleListBegin += i + 1;
  buffer[i] = '\0';

  int32_t count = atoi((char *)buffer);
  assertTrue(count > 0, "TrueTypeAsset::updateAsync: Invalid variant count");

  // Now begin parsing each font style.
  this->state = TRUE_TYPE_ASSET_STATE_READ_VARIANT;
  assetStyles.clear();
  while(assetStyles.size() != count) {
    struct TrueTypeAssetStyle style;

    // Buffer
    this->loader.setPosition(styleListBegin);
    read = this->loader.read(buffer, 32);
    assertTrue(read == 32, "TrueTypeAsset::updateAsync: Could not read variant");

    // Read style
    i = 0;
    while(buffer[i] != ':' && i < 64) i++;
    buffer[i] = '\0';
    style.style = atoi((char *)buffer);
    styleListBegin += i + 1;
    
    // Buffer
    this->loader.setPosition(styleListBegin);
    read = this->loader.read(buffer, 32);
    assertTrue(read == 32, "TrueTypeAsset::updateAsync: Could not read variant style");

    // Read length
    i = 0;
    while(buffer[i] != '|' && i < 64) i++;
    buffer[i] = '\0';
    styleListBegin += i + 1;
    style.dataSize = atol((char *)buffer);

    // Push
    assetStyles.push_back(style);
  }

  // Now we are at the first byte of the first style.
  this->state = TRUE_TYPE_ASSET_STATE_ADJUSTING_OFFSETS;
  auto itStyle = assetStyles.begin();
  while(itStyle != assetStyles.end()) {
    (*itStyle).dataOffset = styleListBegin;
    styleListBegin += (*itStyle).dataSize;
    styleListBegin += 1; // Vertical bar
    ++itStyle;
  }

  // Init FreeType
  this->state = TRUE_TYPE_ASSET_STATE_INIT_FREETYPE;
  int32_t ret = FT_Init_FreeType(&fontLibrary);
  assertTrue(ret == 0, "TrueTypeAsset::updateAsync: Could not init FreeType");

  // Done parsing!
  this->state = TRUE_TYPE_ASSET_STATE_READY;
  this->loaded = true;
}

usagelockid_t TrueTypeAsset::lock(struct TrueTypeFaceTextureStyle style) {
  assertTrue(this->state == TRUE_TYPE_ASSET_STATE_READY, "TrueTypeAsset::lock: Asset is not ready");

  // Try and find an existing texture that matches this style
  auto it = this->textureByStyle.find(style);
  TrueTypeFaceTexture *texture = nullptr;

  if(it == this->textureByStyle.end()) {
    // Does not exist, Find asset style
    auto itAssetStyle = this->assetStyles.begin();
    while(itAssetStyle != this->assetStyles.end()) {
      if((*itAssetStyle).style == style.style) {
        // Found
        break;
      }
      ++itAssetStyle;
    }
    assertTrue(itAssetStyle != this->assetStyles.end(), "TrueTypeAsset::lock: Could not find asset style");

    // Create and read buffer.
    uint8_t *dataBuffer = (uint8_t*)memoryAllocate(sizeof(uint8_t) * itAssetStyle->dataSize);
    this->loader.rewind();
    this->loader.setPosition(itAssetStyle->dataOffset);
    auto read = this->loader.read(dataBuffer, itAssetStyle->dataSize);
    assertTrue(read == itAssetStyle->dataSize, "TrueTypeAsset::lock: Could not read data");

    // Create the face
    FT_Face face;
    auto ret = FT_New_Memory_Face(this->fontLibrary, (FT_Byte*)dataBuffer, itAssetStyle->dataSize, 0, &face);
    assertTrue(ret == 0, "TrueTypeAsset::lock: Could not create face");
    texture = new TrueTypeFaceTexture(face, style);
    memoryFree(dataBuffer);

    this->textures.push_back(texture);
    this->textureByStyle[style] = texture;
  } else {
    // Exists
    texture = it->second;
  }

  auto lock = this->locks.createLock();
  this->textureByLock[lock] = texture;
  this->locksByTexture[texture].push_back(lock);
  return lock;
}

TrueTypeFaceTexture * TrueTypeAsset::getTexture(usagelockid_t id) {
  auto it = this->textureByLock.find(id);
  assertTrue(it != this->textureByLock.end(), "TrueTypeAsset::getTexture: Could not find texture");
  return it->second;
}

void TrueTypeAsset::unlock(usagelockid_t id) {
  this->locks.removeLock(id);
}

TrueTypeAsset::~TrueTypeAsset() {
  auto it = this->textures.begin();
  while(it != this->textures.end()) {
    delete (*it);
    it++;
  }

  FT_Done_FreeType(this->fontLibrary);
}