// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "util/string.hpp"

namespace Dawn {
  static inline std::string rawParser(std::string v, std::string *error) {
    return v;
  };

  static inline std::string stringParser(std::string v, std::string *error) {
    // Replace slashes and quotes
    v = stringReplaceAll(v, "\\", "\\\\\\");
    v = stringReplaceAll(v, "\"", "\\\"");

    // Newlines.
    v = stringReplaceAll(v, "\n", "\\n");

    return "\"" + v + "\"";
  };

  static inline std::string floatParser(std::string v, std::string *error) {
    v = stringTrim(v);

    // Make sure number contains decimal
    if(v.find(".") == std::string::npos) {
      v += ".0";
    }
    // Make sure number contains a number before the decimal
    if(v.find(".") == 0) {
      v = "0" + v;
    }
    // Make sure ends with f
    if(v.find("f") == std::string::npos) {
      v += "f";
    }
    return v;
  };

  static inline std::string intParser(std::string v, std::string *error) {
    v = stringTrim(v);
    return v;
  }

  static inline std::string boolParser(std::string v, std::string *error) {
    v = stringTrim(v);
    if(v == "true") return "true";
    if(v == "false") return "false";
    *error = "Invalid bool value: " + v;
    return std::string("");
  }

  static inline std::string vec2Parser(std::string v, std::string *error) {
    // Split string by comma into two strings that we pass into float
    auto split = stringSplit(v, ",");
    if(split.size() != 2) {
      *error = "Invalid vec2 value: " + v;
      return std::string("");
    }
    return std::string(
      "glm::vec2(" +
        floatParser(split[0], error) + ", " +
        floatParser(split[1], error) +
      ")"
    );
  };


  static inline std::string vec3Parser(std::string v, std::string *error) {
    // Split string by comma into two strings that we pass into float
    auto split = stringSplit(v, ",");
    if(split.size() != 3) {
      *error = "Invalid vec3 value: " + v;
      return std::string("");
    }
    return std::string(
      "glm::vec3(" +
        floatParser(split[0], error) + ", " +
        floatParser(split[1], error) + ", " +
        floatParser(split[2], error) +
      ")"
    );
  };


  static inline std::string vec6Parser(std::string v, std::string *error) {
    // Split string by comma into two strings that we pass into float
    auto split = stringSplit(v, ",");
    if(split.size() != 6) {
      *error = "Invalid vec6 value: " + v;
      return std::string("");
    }
    return std::string(
      "glm::vec6(" +
        floatParser(split[0], error) + ", " +
        floatParser(split[1], error) + ", " +
        floatParser(split[2], error) + ", " +
        floatParser(split[3], error) + ", " +
        floatParser(split[4], error) + ", " +
        floatParser(split[5], error) +
      ")"
    );
  };


  static inline std::string vec4Parser(std::string v, std::string *error) {
    // Split string by comma into two strings that we pass into float
    auto split = stringSplit(v, ",");
    if(split.size() != 4) {
      *error = "Invalid vec4 value: " + v + " (incorrect split len " + std::to_string(split.size()) + ")";
      return std::string("");
    }
    return std::string(
      "glm::vec4(" +
        floatParser(split[0], error) + ", " +
        floatParser(split[1], error) + ", " +
        floatParser(split[2], error) + ", " +
        floatParser(split[3], error) +
      ")"
    );
  };

  static inline std::string colorParser(std::string v, std::string *error) {
    return "Color::fromString(" + stringParser(v, error) + ")";
  };

  static inline std::string uiComponentAlignParser(std::string v, std::string *error) {
    v = stringToLowercase(v);
    if(v.find("left") != std::string::npos) return "UI_COMPONENT_ALIGN_START";
    if(v.find("center") != std::string::npos) return "UI_COMPONENT_ALIGN_MIDDLE";
    if(v.find("right") != std::string::npos) return "UI_COMPONENT_ALIGN_END";
    if(v.find("top") != std::string::npos) return "UI_COMPONENT_ALIGN_START";
    if(v.find("bottom") != std::string::npos) return "UI_COMPONENT_ALIGN_END";
    if(v.find("stretch") != std::string::npos) return "UI_COMPONENT_ALIGN_STRETCH";
    if(v.find("start") != std::string::npos) return "UI_COMPONENT_ALIGN_START";
    if(v.find("middle") != std::string::npos) return "UI_COMPONENT_ALIGN_MIDDLE";
    if(v.find("end") != std::string::npos) return "UI_COMPONENT_ALIGN_END";
    *error = "Invalid UIComponentAlign value: " + v;
    return "";
  }

  static inline std::string uiComponentAlignUnitParser(std::string v, std::string *error) {
    v = stringToLowercase(v);
    if(v.find("scale") != std::string::npos) return "UI_COMPONENT_ALIGN_UNIT_SCALE";
    if(v.find("percent") != std::string::npos) return "UI_COMPONENT_ALIGN_UNIT_PERCENT";
    if(v.find("ratio") != std::string::npos) return "UI_COMPONENT_ALIGN_UNIT_RATIO";
    *error = "Invalid UIComponentAlignUnit value: " + v;
    return "";
  }

  static inline std::string uiLabelTextAlignParser(std::string v, std::string *error) {
    v = stringToLowercase(v);
    if(v.find("left") != std::string::npos) return "UI_LABEL_TEXT_ALIGN_LEFT";
    if(v.find("center") != std::string::npos) return "UI_LABEL_TEXT_ALIGN_CENTER";
    if(v.find("right") != std::string::npos) return "UI_LABEL_TEXT_ALIGN_RIGHT";
    *error = "Invalid UILabelTextAlign value: " + v;
    return "";
  }

  static inline std::function<std::string(std::string, std::string*)> parserFromTypeName(std::string type) {
    std::function<std::string(std::string, std::string*)> parser = rawParser;

    if(type.find("string") != std::string::npos) {
      parser = stringParser;
    } else if(type.find("float") != std::string::npos) {
      parser = floatParser;
    } else if(type.find("Color") != std::string::npos) {
      parser = colorParser;
    } else if(type.find("vec2") != std::string::npos) {
      parser = vec2Parser;
    } else if(type.find("vec3") != std::string::npos) {
      parser = vec3Parser;
    } else if(type.find("vec4") != std::string::npos) {
      parser = vec4Parser;
    } else if(type == "int32_t" || type == "int" || type == "uint32_t" || type == "uint") {
      parser = intParser;
    } else if(type == "bool_t") {
      parser = boolParser;
    } else if(type == "flag_t") {
      parser = rawParser;
    } else if(type.starts_with("enum")) {
      // Handle Enum Cases
      if(type.ends_with("UIComponentAlign")) {
        parser = uiComponentAlignParser;
      } else if(type.ends_with("UIComponentAlignUnit")) {
        parser = uiComponentAlignUnitParser;
      } else if(type.ends_with("UILabelTextAlign")) {
        parser = uiLabelTextAlignParser;
      } else {
        parser = rawParser;
      }
    } else if(type.find("*") == (type.size() - 1)) {
      type = type.substr(0, type.size() - 1);
      parser = rawParser;
    } else {
      throw std::string("Invalid parser type");
    }

    return parser;
  }
}